<?php

/*
	module - base class for all modules
	Copyright (c) 2009
	
	This file is part of NicNames.
	
	NicNames is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as 
	published by the Free Software Foundation, either version 3 of
	the License, or (at your option) any later version.
	
	NicNames is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with NicNames.  If not, see <http://www.gnu.org/licenses/>.
*/

class module
{
	function __construct($templateengine, $format, $fullaccess)
	{
		$this->template = $templateengine;
		$this->format = $format;
		$this->fullaccess = (bool)$fullaccess;
		$this->vars = array();
		$this->context = context::getinstance();
		$this->controls = controls::getinstance();
		$this->realtime = false;
		$this->identities = array();
		$this->maxage = $this->vary = null; // used for caching
		$this->status = null; // set to array of (0 => statuscode, 1 => statustext)
		$this->result = null;
	}

	function denyaccess($description = 'access this page')
	{
		$this->status = array(403, 'Forbidden');
		$this->template->vars['notices'][] = "Sorry, you do not have permission to $description.";
		$this->vars['error'] = 'notauthorized';
	}

	function oai_dcformat()
	{
		$document = new DOMDocument();

		if (is_array($this->identities) && count($this->identities))
		{
			$identity = reset($this->identities);
			$node = $identity->getOAIDC($document);
			$document->appendChild($node);
			if (count($this->identities) > 1)
			{
				$document->appendChild($document->createComment("Note: The \"oai_dc\" format can only represent one record at a time.  Only the first result is supplied in this document."));
			}
			$document->appendChild($document->createComment("Note: The \"oai_dc\" format is not a good format for representing the metadata that NicNames stores.  Please consider using Nicnames XML (\"xml\") or one of the other supported types."));
		}
		return $document->saveXML();
	}

	function xmlformat()
	{
		$document = new DOMDocument();
		$rootnode = $document->createElement("nicnames");
		$rootnode->setAttribute('xmlns', 'http://swin.edu.au/lib/research/nicnames');

		$document->appendChild($rootnode);

		if (is_array($this->vars)) foreach ($this->vars as $var => $val)
		{
			$node = $document->createElement("message", is_scalar($val) ? htmlspecialchars($val) : htmlspecialchars(json_encode($val)));
			$node->setAttribute('code', $var); 
			$rootnode->appendChild($node);
		}

		if (isset($this->template->vars['notices']))
			foreach ($this->template->vars['notices'] as $notice)
		{
			$node = $document->createElement("message", htmlspecialchars((string)$notice));
			$node->setAttribute('code', 'notice');
			$rootnode->appendChild($node);
		}

		if (is_array($this->identities)) foreach ($this->identities as $identity)
		{
			$node = $identity->getXML($document);
			$rootnode->appendChild($node);
		}
		return $document->saveXML();
	}

	function jsonformat()
	{
		$obj = new StdClass;
		$obj->vars = &$this->vars;
		$obj->notices = isset($this->template->vars['notices']) ? $this->template->vars['notices'] : array();
		$result = $this->result;
		if ($result === null)
		{
			$result = array();
			if (is_array($this->identities)) foreach ($this->identities as $id => $identity)
				$result[$id] = $identity->getrendered(65535);
		}
		$obj->result = $result;
		return json_encode($obj);
	}

	protected function restmethod()
		// if the HTTP request is a POST or PUT with a body, or a DELETE, returns
		// the method.  Returns null otherwise.  Ordinary POST with $_POST vars is
		// not counted.
	{
		if (!empty($_POST) || empty($_SERVER['REQUEST_METHOD'])
			|| !in_array($_SERVER['REQUEST_METHOD'], array('PUT', 'POST', 'DELETE')))
			return null;
		if ($_SERVER['REQUEST_METHOD'] != 'DELETE' && !empty($_SERVER['CONTENT_LENGTH'])
			&& $_SERVER['CONTENT_LENGTH'] == 0)
			return null;
		return $_SERVER['REQUEST_METHOD'];
	}

	protected function getxmlidentities($filename)
		// reads xml input from $filename and returns an array of identities from it
		// or empty array if it contained no identities
		// or null if it was invalid or couldn't be opened
	{
		$validxml = false;
		$doc = new DomDocument();
		if (!$doc->load($filename, LIBXML_NOWARNING | LIBXML_NOERROR)) return null;
		if (!($root = $doc->documentElement)) return null;

		require_once(RESOURCE_DIR . 'includes/identity.inc.php');

		$identities = array();
		for ($i = 0, $len = $root->childNodes->length; $i < $len; $i++)
		{
			$node = $root->childNodes->item($i);
			if ($node->nodeType != XML_ELEMENT_NODE || $node->tagName != 'identity')
				continue;

			$identity = new identity();
			$identity->setfromXML($node);
			$errors = $identity->validate();
			if (!empty($errors)) return null;
			$identities[] = $identity;
		}
		return $identities;
	}

	protected function findidentities($perpage = 20, $cookieremember = true, $rendernum = 4)
	// finds identities matching the given "start" and "query" parameters
	// populates template variables, allowing the results to be shown
	// returns identities, or null if no search was made, or empty array
	// if no results found
	{
		// when making a query, the query string is stored in a cookie for later
		// use.  this is a session cookie only.
		$query = $this->context->load_var('query', 'REQUEST', 'string');
		if ($cookieremember)
		{
			if ($query === null)
				$query = $this->context->load_var('action_query', 'COOKIE', 'string');
			else
				$this->context->setcookie('action_query', $query);
		}

		$this->template->vars['action_query'] = $this->vars['querystring'] = $query;
		$this->querystring = $query;
		if ($query == '') return null;

		$indexlookup = $this->controls->getcontrol('indexlookup');
		$ids = $indexlookup->grammaticalmatch($query);

		return $this->layoutidentities($ids, $perpage, $rendernum);
	}

	protected function layoutidentities($ids, $perpage = 20, $rendernum = 4)
		// for laying out a list of identities in pages.  used by findidentities, but
		// can be used directly if you have a list of ids
	{
		$perpagein = $this->context->load_var('perpage', 'REQUEST', 'int');
		if ($perpagein) $perpage = $perpagein;
		$perpage = min(max($perpage, 1), 1000);

		$rendernumin = $this->context->load_var('rendernum', 'REQUEST', 'int');
		if ($rendernumin) $rendernum = $rendernumin;
		$rendernum = min(max($rendernum, 1), 65535);

		$this->template->vars['action_count'] = $this->vars['resultcount'] = (int)count($ids);

		$start = $this->context->load_var('start', 'REQUEST', 'int');
		$slice = array_slice($ids, $start, $perpage);

		$fetchidentities = $this->controls->getcontrol('fetchidentities');
		$identities = $fetchidentities->getidentities($slice);

		$idfacts = array();
		foreach ($identities as $identity)
		{
			if (!$this->fullaccess) $identity->stripprivate();
			$idfacts[$identity->id] = $identity->getrendered($rendernum);
		}

		$this->template->vars['action_resultstart'] = $this->vars['resultstart'] = (int)$start;
		$this->template->vars['action_resultperpage'] 
			= $this->vars['resultperpage'] = (int)$perpage;

		$this->template->vars['action_identities'] = $this->resultidentities = $identities;
		$this->template->vars['action_facts'] = $idfacts; 

		return $identities;
	}

	function chooseperson($fieldname = null, $required = false, $url, $description = 'person')
	{
		if (!$this->fullaccess) return $this->denyaccess('choose a relation person');
		$this->findidentities(10, false);

		$postvars = array();
		foreach ($_POST as $key => $val) if ($key != "field_$fieldname")
			$postvars[$key] = $this->context->load_var($key, 'POST', 'mixed');

		$this->template->vars['action_fieldname'] = $fieldname;
		$this->template->vars['action_required'] = $required;
		$this->template->vars['action_postvars'] = $postvars;
		$this->template->vars['action_url'] = $url;
		$this->template->vars['action_description'] = $description;

		return $this->template->runtemplate('common_chooseperson');
	}

	protected function validateobject($obj, $fields, $input = true, $prefix = 'field_')
	// validators object $obj according to fields $fields - the format if $fields is
	// as in relation->getfields()
	// if $input is true, then this also reads in the value from input.  otherwise
	// it just validates the object as is.
		// returns an empty array if there are no validation errors, otherwise contains
		// an array of ($fieldname => $errormsg)
	{
		$errors = array();

		foreach ($fields as $field)
		{
			list($name, $type, $label, $desc, $maxlen, $required) = $field;
			$options = isset($field[6]) ? $field[6] : array();
			$value = isset($obj->{$name}) ? $obj->{$name} : null;

			// begin input
			if ($input)
			{
				$new = $this->context->load_var($prefix.$name,'REQUEST',$type,array_keys($options));
				$rawnew = $this->context->load_var($prefix.$name,'REQUEST','mixed');

				if ($rawnew !== null)
				{
					$value = $new;
					if ($new === null && $rawnew != '')
					{
						$errors[$name] = "The value given for the $label field was not valid.";
						continue;
					}
				}

				if (!empty($options['identitytype']))
				{
					if ($new != '')
					{
						$fetchidentities = $this->controls->getcontrol('fetchidentities');
						$identities = $fetchidentities->getidentities(array($new), 0);
						if (empty($identities))
						{
							$errors[$name] = "The value given for the $label field was not a valid identity.  This value must be an identity in NicNames.";
							continue;
						}
						$value = current($identities);
					} 
					else $value = null;
				}
			}


			$emptyval = is_object($value) ? $value === null : strlen($value) == 0;

			// post input validation
			if ($required && $emptyval && $type != 'yesno')
			{
				$errors[$name] = "The $label field is necessary.";
				continue;
			}

			if (!is_object($value))
			{
				if ($maxlen && strlen($value) > $maxlen)
				{
					$errors[$name] = "The $label field is too long.  The maximum length is approx. " . number_format($maxlen) . " characters.";
					continue;
				}

				if (!empty($options['min']) && strlen($value) && $value < $options['min'])
				{
					$errors[$name] = "The value for $label was too low.  The lowest allowed value is " . $options['min'] . '.';
					continue;
				}

				if (!empty($options['max']) && strlen($value) && $value > $options['max'])
				{
					$errors[$name] = "The value for $label was too high.  The highest allowed value is " . $options['max'] . '.';
					continue;
				}
			}

			if (!empty($options['identitytype']) && $value !== null && !($value instanceof identity))
			{
				$errors[$name] = "The value for $label is not a valid identity.";
				continue;
			}

			$obj->{$name} = $value; 

		}
		return $errors;

	}

	function startrealtime($contenttype = 'text/html; charset=utf-8', $status = '200 OK')
	{
		if (function_exists('apache_setenv')) apache_setenv('no-gzip', 1);
    ini_set('zlib.output_compression', 0);
		for ($i = ob_get_level(); $i > 0; $i--) @ob_end_clean();
		header("HTTP/1.1 $status");
		header("Content-type: $contenttype");
		$this->realtimebytes = 0;
		$this->realtime = true;
	}

	protected function getbaseurl()
	{
		if (empty($_SERVER['SERVER_NAME'])) return "http://example.com/unknown-base-uri";
		$protocol = (!empty($_SERVER['HTTPS']) && $_SERVER['HTTPS'] != "off") ? "https" : "http";
		$host = $_SERVER['SERVER_NAME'];
		$port = ":" . $_SERVER['SERVER_PORT'];
		if ($port == ($protocol == 'https' ? ":443" : ":80")) $port = "";
		list($uri) = explode('?', $_SERVER['REQUEST_URI']);
		return $protocol . "://" . $host . $port . $uri;
	}

	protected function getnicnamesurl()
	{
		$url = $this->getbaseurl();
		if ($this->fullaccess) $url = preg_replace('#[^/]+/([^/]*)$#', '$1', $url);
		return $url;
	}

	function outputrealtime($data)
	{
		$this->realtimebytes += strlen($data);
		echo $data;
		flush();
	}
}

?>
