<?php

/*
	name - a generic class for transforming and comparing people's names
	Copyright (c) 2009
	
	This file is part of NicNames.
	
	NicNames is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as 
	published by the Free Software Foundation, either version 3 of
	the License, or (at your option) any later version.
	
	NicNames is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with NicNames.  If not, see <http://www.gnu.org/licenses/>.
*/

class name
{
	function name($textname = null)
	{
		if ($textname) $this->extract($textname);
	}
	
	function getparts()
	// returns array of name parts
	{
		return array(
			'surname' => $this->surname,
			'givennames' => $this->givennames,
			'title' => $this->title,
			'extra' => $this->extra
			);
	}
	
	function getstandardname()
	{
		return $this->getmappedname(false, true, 2, 0x01, false, false);
	}
	
	function getmappedname($showtitle = true, $reversed = true, $showgivens = 8, $expandinitials = 0xff, $showextra = true, $surnamefirst = false)
	// $reversed will output "Smith, Tony John, Dr" instead of "Tony John Smith"
	// $surnamefirst will output "Dr Smith Tony John" instead of "Tony John Smith" (useful for cultural differences)
	// $expandinitials is a bitfield specifying which names will be expanded
	// (if we have the expanded form stored)
	// 0: "T. J. W. Smith" (no names expanded)
	// 1: "Tony J. W. Smith" (1st name expanded)
	// 2: "T. John W. Smith" (2nd name expanded)
	// 3: "Tony John W. Smith" (1st and 2nd names expanded)
	// 4: "T. J. Walter Smith" (3rd name expanded)
	// 5: "Tony J. Walter. Smith" (1st and 3rd names expanded)
	// 6: "T. John Walter. Smith" (2nd and 3rd names expanded)
	// 7: "Tony John Walter Smith" (1st, 2nd and 3nd names expanded)
	// 0xffff: "Tony John Walter Robert Malcom Jeremy Smith" (all names expanded)
	// $showgivens is the maximum number of given names to show (8 is the max)
	// $showtitle shows the title eg "Dr", "Prof"
	{
		// initialise
		$givens = $this->givennames;
		$init = $expandinitials;
		$numg = $showgivens;
		foreach ($givens as $id => $given)
		{
			if ($numg-- <= 0)
			{
				unset($givens[$id]);
				continue;
			}
			if (!($init & 1))
				$givens[$id] = $this->nametoinitial($given) . '.';
			$init >>= 1;
		}
		$textgivens = implode(' ', $givens);
		$surname = $this->surname;
		$title = $this->title;
		if ($textgivens)
		{
			$textname = $reversed ?
				"$surname, $textgivens" :
				($surnamefirst ? "$surname $textgivens" : "$textgivens $surname");
		}
		else $textname = $surname;
		if ($showtitle && $title) $textname = "$title $textname";
		$extra = $this->extra;
		if ($showextra && $extra) $textname = "$textname $extra";
		return $textname;		
	}
	
	// private
	function alpha26($name)
	{
		$name = strtr($name, "", "SOZsozYYuAAAAAAACEEEEIIIIDNOOOOOOUUUUYsaaaaaaaceeeeiiiionoooooouuuuyy");
		$name = preg_replace('#[^a-zA-Z]+#', '', $name);
		return strtoupper($name);
	}
	
	// public
	function getpid($number = 0)
	// gets a PID (unique person identifier) based on this name
	// to ensure uniqueness, call it with an increasing number (beginning 0)
	// until it returns one that is not taken
	{
		$surn = substr($this->alpha26($this->surname), 0, 4);
		$givn = substr($this->alpha26(reset($this->givennames)) ,0, 2);
		$pid = $surn . $givn;
		while (strlen($pid) < 6) $pid .= "0";
		if ($number > 0)
		{
			$numstring = "$number";
			if (strlen($numstring) > 6)
				trigger_error("Exhausted 1 million possible unique person identifiers", E_USER_ERROR);
			$pid = substr($pid, 0, 6 - strlen($numstring)) . $numstring;
		} 
		return strtoupper($pid);
	}
	
	// private
	function nametoinitial($name)
	{
		$prefixes = array('de ', 'of ', 'au ');
		foreach ($prefixes as $pref)
		{
			$patt = str_replace(' ', '\s+', $pref);
			if (preg_match("#^$patt#i", $name, $matches))
				$name = substr($name, strlen($matches[0]));
		}
		return strtoupper($name[0]);
	}
	
	// public
	function setname($surname, $givennames = null, $title = null, $extra = null)
	// name won't be translated, so it must be in normal form already, ie
	// capitalisations and periods indicating initials, etc
	{
		$this->title = $title;
		$this->surname = $surname;
		$this->givennames = preg_split('#\s+#', $givennames);		
		$this->extra = $extra;
	}
	
	// private
	function clear()
	{
		$this->surname =
			$this->givennames =
			$this->title =
			$this->extra = 
			null;
	}
	
	// public
	function extract($textname)
	// extracts a name from a name given in plain text
	{
		$this->clear();
		
		$textname = $this->extractextra($textname);
		$textname = $this->extracttitle($textname);
				
		// do reversal
		$reversalparts = preg_split('#\s*,\s*#', $textname);
		foreach ($reversalparts as $id => $val) $reversalparts[$id] = trim($val);
		$textname = implode(" ", array_reverse($reversalparts));
		// name is now in forward order
		if (!$this->title) $textname = $this->extracttitle($textname);
		$nameparts = preg_split('#\s+#', trim($textname));
		if (!count($nameparts)) return;
		$surname = array_pop($nameparts);
		while ($this->isprefix(end($nameparts)))
			$surname = array_pop($nameparts) . " " . $surname;
		$this->namemassage($surname);
		$this->surname = $surname;
		array_walk($nameparts, array($this, 'namemassage'));
		$this->givennames = $nameparts; 		
	}
	
	// private
	function isprefix($name)
	{
		$prefixes = array('de', 'of', 'van', 'der', 'ten');
		if (in_array(trim(strtolower($name)), $prefixes)) return true;
		return false;
	}
	
	// private
	function namemassage(&$textname)
	{
		$textname = trim($textname);
		$textname = preg_replace('#[()";:\[\]]#', '', $textname);
		$textname = preg_replace('#[_\s]+#', ' ', $textname);
		if (preg_match('#^(\w\.|[b-hj-tv-xz])$#', $textname))
		{
			// initialisation
			$textname = strtoupper($textname[0] . '.');
			return;
		}
		$textname = str_replace('.', '', $textname);
		$textname = strtolower($textname);
		if (preg_match_all('#(?<=^|\b|^mc|^mac)\w#i', $textname, $matches, PREG_OFFSET_CAPTURE))
		{
			foreach ($matches[0] as $val)
			{
				$offset = $val[1];
				list($word) = explode(' ', substr($textname, $offset));
				if (!$this->isprefix($word))
					$textname = substr($textname, 0, $offset) . strtoupper($textname[$offset]) . substr($textname, $offset+1);
			}
		}
		return;
	}
	
	// private
	function extractextra($textname)
	// gets extra information from the end of a name such as date of birth/death,
	// information in brackets etc
	// returns the $textname with the extra information removed, if any was
	// found
	// \s*-?\s*(\d{2,4}|or\s+\d{2,4})?\.?
	{
		if (preg_match('#,?\s+((,?\s*(([Cc]a?\.?)?\s*\d{4}[^\[\(,]*|\[[^\]]*\]|\([^\)]*\)|i\.?\s?e\.[^\[\9,]*|[dbf]\.?\s*\d{4}\.?))+)\s*$#', $textname, $matches, PREG_OFFSET_CAPTURE))
		{
			$this->extra = trim($matches[1][0]);
			$textname = substr($textname, 0, $matches[1][1]);
		}
		
		return $textname;
	}
	
	// private
	function getphon($input)
	// returns a phoenetic representation of the text
	{
		static $cache = array();
		if (isset($cache[$input])) return $cache[$input];
		$text = $input;
		
		$text = strtr($text, "", "SOZsozYYuAAAAAAACEEEEIIIIDNOOOOOOUUUUYsaaaaaaaceeeeiiiionoooooouuuuyy");
		$text = strtolower(trim(preg_replace('#[^a-zA-Z]+#', '', $text)));
		$replace = array(
			'qu' => 'kw', 'kn' => 'n', 'gn' => 'n', 'lm' => 'm',
			'ld' => 'd', 'ce' => 'se', 'psy' => 'sy', 'wr' => 'r',
			'ing' => 'in', 'ich' => 'itch', 'own' => 'on', 'gua' => 'gwa',
			'igh' => 'i', 'ph' => 'f', 'cy' => 'sy', 'tion' => 'shon',
			'isch' => 'ish', 'schm' => 'shm', 'ough' => 'oe',
			'dge' => 'je', 'dgy' => 'jy', 'dge' => 'je', 'xx' => 'x',
			// foreign words
			'aux' => 'au',
			// single letters
			'x' => 'ks',
			);
		$text = str_replace(array_keys($replace), array_values($replace), $text);
		$text = preg_replace('#le$#', 'el', $text);
		$text = preg_replace('#([aeiouy][^aeiuoyh]{1,3})e(?![rmlpdstz])#', '\\1', $text);
		$text = preg_replace('#(?<!^m)(?<!^ma)ci#', 'si', $text); // ci => si except mcinnes
		$text = preg_replace('#ew(?![aeiouy])#', 'u\\1', $text);
		$text = preg_replace('#sh([aeiouy])#', 'j\\1', $text);
		$text = preg_replace('#^mc([^aeouy])#', 'mac\\1', $text);
		$text = preg_replace('#([aeiouy])w#', '\\1', $text);
		$text = preg_replace('#([aeiouy])r([^aeiouy])#', '\\1\\2', $text);
		$text = preg_replace('#([aeiouy])r$#', '\\1', $text);
		// === groups:
		// o: aeiouy
		// m: bmnp
		// k: cgkqxj
		// t: dstz
		// f: fvw
		// r: lr
		// unused: h		
		$tr_from =   'abcdefghijklmnopqrstuvwxyz';
		$tr_to =     'omktofk okkrmmomkrttoffkot';
		$text = strtr($text, $tr_from, $tr_to);
		$text = str_replace(' ', '', $text);
		$text = preg_replace('#([a-z])\\1+#', '\\1', $text);
		$cache[$input] = $text;
		if (count($cache) > 250) $cache = array_slice($cache, -125);
		return $text;
	}
	
	function getindexes()
	// gets phonetic indexes in associative array
	{
		$sur_phon = $this->getphon($this->surname);
		$indexes = array();
		$indexes['i_surn'] = $this->getalphaindex($this->surname);
		$indexes['i_given'] = $this->getalphaindex(implode(' ', $this->givennames));
		$indexes['i_phon'] = $this->getalphaindex($sur_phon);
		$indexes['i_phon_r'] = $this->getalphaindex(strrev($sur_phon));
		return $indexes;		
	}
	
	//private
	function comparestrings($s1, $s2)
	// returns a score indicating the difference of the strings
	// 0 = equal, 100 = very different
	// takes into account similar spellings and similar pronunciations
	// not case sensitive
	// try to limit the string lengths to around 20 chars each
	// this is an O(m*n)
	{
		$s1 = strtolower(trim($s1));
		$s2 = strtolower(trim($s2));
		static $cache = array();
		if (isset($cache[$s1][$s2])) return $cache[$s1][$s2];
		if ($s1 == $s2) return 0;
		$speldis = levenshtein($s1, $s2);
		$phondis = levenshtein($this->getphon($s1), $this->getphon($s2));
		$dis = $speldis + ($phondis * 4.23);
		$len = strlen($s1) + strlen($s2);
		if ($len > 12) $dis /= ($len / 12);
		$value = round($dis * 5);
		$cache[$s1][$s2] = $value;
		if (count($cache, COUNT_RECURSIVE) > 250) $cache = array();
		return $value;
	}
	
	function compareto($name2)
	// calculates score indicating how different this name is to the other
	// takes similar spelling and pronunciation into account
	// the closer to zero, the surer this algorithm is that the names are the same
	// this algorithm is fairly processor intensive
	// 0-22: most likely the same name eg same first name and surname
	// 23-32: looks like same name eg same surname and first initial
	// 33-50: could be the same except for spelling or small variations 
	// 51-100: doesn't look like a match, could be relevant though
	// 101-: doesn't look relevant 
	{
		$parts2 = $name2->getparts();
		
		// surname
		$dif = $this->comparestrings($this->surname, $parts2['surname']);
		$bon = in_array($parts2['surname'], $this->givennames)
			||   in_array($this->surname, $parts2['givennames']);
		$score = $bon ? min(40, $dif) : $dif;		
		
		// givennames
		$given1 = $this->givennames;
		$given2 = $parts2['givennames'];
		$scorevals = array(50, 30, 10, 8, 6, 4, 2, 1);
		while (!empty($scorevals))
		{
			$name1 = array_shift($given1);
			$name2 = array_shift($given2);
			$scoreval = array_shift($scorevals);
			$nscore = 0;
			if (!strlen($name1) || !strlen($name2)) $nscore = 35;
			else
			{
				$initials = (preg_match('#\.$#', $name1) || preg_match('#\.$#', $name2));				if (!$initials)
					$nscore = $this->comparestrings($name1, $name2);
				else
					$nscore = 20;	
				if ($name1{0} != $name2{0}) $nscore += 50;
			}
			$score += round(($nscore * $scoreval) / 100);
		}
		
		// title
		if ($this->title != $parts2['title'])
		{
			if (!strlen($this->title) || !strlen($parts2['title'])) $score += 2;
				else $score += 6;
		}
		
		// extra
		if ($this->extra != $parts2['extra'])
		{
			$extra1 = trim(preg_replace('#[^a-z0-9]#', ' ', strtolower($this->extra)));
			$extra2 = trim(preg_replace('#[^a-z0-9]#', ' ', strtolower($parts2['extra'])));
			$extra1 = substr($extra1, 0, 32);
			$extra2 = substr($extra2, 0, 32);
			if (!strlen($extra1) || !strlen($extra2)) $score += 16;
				else $score += levenshtein($extra1, $extra2) * 3;
		}
		
		return $score;

	}
	
	function getalphaindex($text, $unsigned = false)
	// generates a 32 bit integer from the first few characters (max 7) of the
	// text to
	// aid sorting and fetching strings that are alphabetically nearby
	// lower numbers = earlier in the alphabet
	// if $unsigned is false, it avoids using the highest bit in case it is
	// interpreted as a signing bit (which would affect sorting)
	{
		$value = 0;
		$shift = $unsigned ? 27 : 26;
		$text = preg_replace('#[^a-z]+#', '', strtolower($text));
		$t = substr($text, 0, 7);
		for ($i = 0; $i < 7; $i++)
		{
			if ($i > strlen($t) - 1)
				$cod = 0;
			else
				$cod = ord($t{$i}) + 1 - ord('a');
			$cod = $shift >= 0 ? ($cod << $shift) : ($cod >> (0 - $shift));
			$value |= $cod;
			$shift -= 5;
		}
		return $value;
	}
	
	// private
	function extracttitle($textname)
	// gets the "title" from the start of the text based name ie Ms, Dr, etc
	// returns the name - with the title removed if one was found
	{
		$acceptedtitles = array(
			"A/Prof" => array('aprof', 'a/prof', 'apro', 'a/pro', 'associate professor', 'a/professor', 'a/p', 'a prof', 'a pro'),
			"Adj/Prof" => array('adjprof', 'adj/prof', 'adjpro', 'adj/pro', 'adjunct professor', 'adj prof', 'adj pro'),
			"Em/Prof" => array('emprof', 'em/prof', 'empro', 'em/pro', 'em prof', 'emeritus professor', 'em professor', 'professor emeritus'),
			"Prof Dr" => array('prof dr', 'professor doctor', 'professor dr', 'prof/dr'),
			"Prof" => array('prof', 'professor', 'pro'),
			"Dr" => array('dr', 'doctor', 'ph.d'),
			"Miss" => array('miss'),
			"Mr" => array('mr', 'mister'),
			"Mrs" => array('mrs', 'missus'),
			"Ms" => array('ms'),
			"Herr" => array('herr'),
			"Monsieur" => array('monsieur'),
			"Hr" => array('hr'),
			"Frau" => array('frau'),
			"Mstr" => array('mstr', 'master'),
			"Sir" => array('sir')
			);
		foreach ($acceptedtitles as $title => $permutations)
		{
			foreach ($permutations as $perm)
			{
				$patt = str_replace(' ', '\.?:?\s+', $perm);
				$patt = str_replace('.', '\.', $patt);
				if (preg_match("#^\s*$patt\.?:?\s+#i", $textname, $matches))
				{
					$textname = substr($textname, strlen($matches[0]));
					$this->title = $title;
					return $textname;
				}
			}
		} 
		return $textname;
	}	
}

/*
$name1 = new name("Smith, Lee");
$name2 = new name("Smith, Leigh");

var_export($name1->getparts());
echo "\n";
var_export($name1->getindexes());
echo "\n";
var_export($name2->getindexes());
echo "\n";
var_export($name1->compareto($name2));
 */

?>
