<?php

/*
	relation - base class for a piece of information related to an identity
	Copyright (c) 2009
	
	This file is part of NicNames.
	
	NicNames is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as 
	published by the Free Software Foundation, either version 3 of
	the License, or (at your option) any later version.
	
	NicNames is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with NicNames.  If not, see <http://www.gnu.org/licenses/>.
*/

require_once(RESOURCE_DIR . 'includes/textfuzzy.inc.php');
require_once(RESOURCE_DIR . 'includes/probability.inc.php');

class relation
{
	public $comment = null;
	public $incidence = 1;
	public $score = 1;
	public $verified = false;
	public $startyear = null;
	public $endyear = null;

	protected function getcommonfields()
	{
		return array(
			array('incidence', 'int', 'occurrences', 'How many times has this person been associated with this detail?', 11, true, array('min' => 1, 'max' => 9999, 'default' => 1, 'canhide' => true)),
			array('startyear', 'int', 'start year', 'What year did this start or was the first year it occurred, if applicable?', 4, false, array('min' => 1400, 'max' => '2099', 'default' => date('Y'), 'canhide' => true)),
			array('endyear', 'int', 'end year', 'What year did this end or was the last year this occurred, if applicable?', 4, false, array('min' => 1400, 'max' => '2099', 'default' => date('Y'), 'canhide' => true)),
			array('verified', 'yesno', 'verified?', 'Can you confirm this information is true or from a trusted source?', 1, true),
			array('comment', 'string', 'comment', 'Add an optional comment', 65535, false, array('canhide' => true, 'size' => 128)),
			);
	}

	function validate()
	// checks this relation to see if all its properties are valid according to its field definitions
	// returns array of errors specifying what the error was for each field.  an empty array means
	// everything validated successfully
	{
		$fields = $this->getfields();
		foreach ($fields as $field)
		{
			list($property, $type, $label, $description, $maxlen, $required) = $field;
			$options = (isset($field[6])) ? $field[6] : array();

			// TODO finish

		}
	}

	protected function getcommonpropertiesXML($el)
	// loads common properties into this element
	{
		if ($this->comment != '') $el->setAttribute('comment', $this->comment);
		if ($this->incidence != 1) $el->setAttribute('incidence', $this->incidence);
		if ($this->score != 1) $el->setAttribute('score', $this->score);
		if ($this->verified) $el->setAttribute('verified', 'true');
		if ($this->startyear) $el->setAttribute('startyear', (int)$this->startyear);
		if ($this->endyear) $el->setAttribute('endyear', (int)$this->endyear);
		return $el;
	}

	protected function setcommonfromXML($el)
	// populate this object with properties from the supplied DOM XML element
	{
		if ($el->hasAttribute('comment')) $this->comment = $el->getAttribute('comment');
		if ($el->hasAttribute('incidence')) $this->incidence = $el->getAttribute('incidence');
		if ($el->hasAttribute('score')) $this->score = $el->getAttribute('score');
		if ($el->hasAttribute('verified')) $this->verified = strtolower($el->getAttribute('verified')) == 'true' ? true : false;
		if ($el->hasAttribute('startyear')) $this->startyear = $el->getAttribute('startyear');
		if ($el->hasAttribute('endyear')) $this->endyear = $el->getAttribute('endyear');
	}

	function commonsimilarityfactor($other)
		// compares the common factors of this relation and the other relation and
		// outputs an array($score, $weight) where score is 0 to 1 and weight is relative
		// to 1
	{
		$factors = array();

		if ($this->comment != '' && $other->comment != '')
		{
			$fuzzycomment1 = new textfuzzy($this->comment);
			$fuzzycomment2 = new textfuzzy($other->comment);
			$factors['comment'] = array($fuzzycomment1->similarityto($fuzzycomment2), 1.5);
		}

		if (($this->startyear || $this->endyear) && ($other->startyear || $other->endyear))
		{
			$start1 = $this->startyear ? $this->startyear : 0;
			$end1 = $this->endyear ? $this->endyear : 9999;
			$start2 = $other->startyear ? $other->startyear : 0;
			$end2 = $other->endyear ? $other->endyear : 9999;
			$factors['yearrange'] = array($start1 <= $end2 && $start2 <= $end1, 2.7);
		}

		return array(probability::combinedcorrected($factors), 200);
	}

	function getplaintext()
		// gets a plain text rendition of the content of this relation.  suitable
		// where you need to describe the whole relation in a single text string.
	{
		$strings = $this->getstrings();

		$output = isset($strings['title']) ? $strings['title'] : "";
		if (isset($strings['subtitle']) && $strings['subtitle'] != "")
			$output .= (!empty($strings['comma']) ? ", " : " ") . $strings['subtitle'];
		if (isset($strings['pre-qualifier']) && $strings['pre-qualifier'] != "")
			$output = ucfirst($strings['pre-qualifier']) . ": " . $output;
		if (isset($strings['post-qualifier']) && $strings['post-qualifier'] != "")
			$output .= ", " . $strings['post-qualifier'];

		$bracketed = array();

		if (!empty($strings['incidence']) && $strings['incidence'] != 1) 
			$bracketed[] = $strings['incidence'] . " times";

		$years =
			(!empty($strings['startyear']) ? $strings['startyear'] : "") . "-" .
			(!empty($strings['endyear'])  ? $strings['endyear']   : "");
		if ($years != "-") $bracketed[] = $years;

		if (count($bracketed))
			$output .= " (" . implode(", ", $bracketed) . ")";

		return $output;
	}

	function getcommonstrings()
		// returns array of string fields suitable for human-readable display
	{
		$str = array();
		if ($this->comment != '') $str['comment'] = $this->comment;
		if ($this->startyear !== null) $str['startyear'] = $this->startyear;
		if ($this->endyear !== null) $str['endyear'] = $this->endyear;
		if ($this->incidence != 1) $str['incidence'] = $this->incidence;
		if ($this->id) $str['id'] = $this->id;
		if ($this->verified) $str['verified'] = true;
		$str['type'] = get_class($this);
		return $str;
	}

	function gettextfields()
		// returns array of strings and their weights, suitable for indexing.
		// overload this if needed
	{
		$weights = array(
			'name' => 1,
			'identifier' => 0.98,
			'classification' => 0.55,
			'person' => 0.75,
			'resource' => 0.8,
			'organisation' => 0.5,
			'event' => 0.65,
			);
		$weight = $weights[get_class($this)];
			
		$strings = $this->getstrings();
		$fields = array();
		if (isset($strings['title'])) $fields[] = array($strings['title'], ceil($weight * 120));
		if (isset($strings['subtitle'])) $fields[] = array($strings['subtitle'], ceil($weight * 80));
		if (isset($strings['pre-qualifier'])) $fields[] = array($strings['pre-qualifier'], ceil($weight * 8));
		if (isset($strings['post-qualifier'])) $fields[] = array($strings['post-qualifier'], ceil($weight * 6));
		if (isset($strings['href'])) $fields[] = array($strings['href'], ceil($weight * 2));
		if (isset($strings['comment'])) $fields[] = array($strings['comment'], ceil($weight * 8));
		if (isset($strings['startyear'])) $fields[] = array($strings['startyear'], ceil($weight * 5));
		if (isset($strings['endyear'])) $fields[] = array($strings['endyear'], ceil($weight * 3));
		if (isset($strings['incidence'])) $fields[] = array($strings['incidence'], ceil($weight * 1));

		return $fields;
	}

	function merge($other)
		// merges another object with this object.  this should be done only when
		// it is quite certain they are duplicates, as if any properties of the
		// incoming object are non-null and conflict with non-null existing
		// properties, then it will only keep one of the two, which is usually
		// the local one.
		// when overloading this, remember to call the parent at some stage, 
		// because this does some special stuff with the common properties
	{
		foreach ($other as $name => $value)
		{
			if (
				(!isset($this->$name))
				|| ($name === 'incidence' && $value > $this->incidence)
				|| ($name === 'startyear' && $value < $this->startyear)
				|| ($name === 'endyear' && $value > $this->endyear)
				|| ($name === 'verified' && $value > $this->verified)
				|| (is_string($value) && is_string($this->$name) && strlen($value) > strlen($this->$name))
				|| (empty($this->$name) && !empty($value))
			)
				$this->$name = $value;
		}
	}
}

?>
